from datatype_writer.link import Link, HexLink, ByteLink


def hex_dump(byte_array, endianess='be'):
    str_array = [hex(b)[2:].zfill(2).upper() for b in byte_array]
    if endianess == 'le':
        str_array = str_array[::-1]
    return " ".join(str_array)


class Format:
    def __init__(self, data, endianess='be'):
        self.data = data
        self.endianess = endianess

    def get_bytes(self):
        pass

    def get_hex_dump(self):
        pass

    def __and__(self, other):
        return HexLink(self.get_hex_dump() + " " + other.get_hex_dump())

    def __or__(self, other):
        return ByteLink(self.get_bytes() + other.get_bytes())

    def __add__(self, other):
        return Link(self.get_hex_dump() + " " + other.get_hex_dump(), self.get_bytes() + self.get_bytes())

    def __str__(self):
        return self.get_hex_dump()

    def __repr__(self):
        return f'Format({self.data},{self.endianess})'

    def set_data(self, value):
        self.data = value




from datatype_writer.format import Format


def hex_dump(byte_array, endianess):
    str_array = [hex(b)[2:].zfill(2).upper() for b in byte_array]
    if endianess == 'le':
        str_array = str_array[::-1]
    return " ".join(str_array)


def convert_bytes(string, encoding, size):
    bytes_array = bytes(string, encoding)
    if size == -1:
        return bytes_array
    elif (size - len(bytes_array)) >= 0:
        difference = size - len(bytes_array)
        return bytes(difference) + bytes_array
    else:
        return bytes_array[0:size]


class String(Format):
    def __init__(self, data, endianess, size, encoding):
        Format.__init__(self, data, endianess)
        self.size = size
        self.encoding = encoding

    def get_bytes(self):
        byte_array = bytes(self.data, self.encoding)
        if self.endianess == 'le':
            byte_array = byte_array[::-1]
        return byte_array

    def get_hex_dump(self):
        byte_array = bytes(self.data, self.encoding)
        dump = hex_dump(byte_array, self.endianess)
        return dump

    def __repr__(self):
        return f'String({self.data},{self.endianess},{self.size},{self.encoding})'


class Unicode(String):
    def __init__(self, data, endianess='be', size=-1, encoding='utf-8'):
        String.__init__(self, data, endianess, size, encoding)



from datatype_writer.format import Format
from math import ceil


class Bit(Format):
    def __init__(self, val, size, endianess='be'):
        Format.__init__(self, None, "big" if endianess == 'be' else 'little')
        self.bits = None
        self.size = size
        if val is not None:
            self.set_data(val)
        else:
            self.data = None

    def get_bytes(self):
        return int(self.bits, 2).to_bytes(length=max(ceil(self.size / 8), 1), byteorder=self.endianess)

    def get_hex_dump(self):
        byte_array = self.get_bytes()
        hex_dump = [hex(b)[2:].zfill(2) for b in byte_array]
        return " ".join(hex_dump).upper()

    def __sub__(self, other):
        return Bit(self.bits + other.bits, self.size + other.size)

    def set_data(self, value):
        if type(value) == int:
            self.bits = bin(value)[2:].zfill(self.size)
        elif type(value) == str:
            self.bits = value

    def __repr__(self):
        return f"Bit({self.bits}{self.size}{self.endianess})"


from datatype_writer.format import Format


class Integer(Format):
    def __init__(self, data, endianess='be', size=1, signed=False):
        Format.__init__(self, data,  "big" if endianess == 'be' else 'little')
        self.size = size
        self.signed = signed

    def get_bytes(self):
        return int(self.data).to_bytes(length=self.size, byteorder=self.endianess, signed=self.signed)

    def get_hex_dump(self):
        byte_array = self.get_bytes()
        hex_dump = [hex(b)[2:].zfill(2) for b in byte_array]
        return " ".join(hex_dump).upper()

    def __repr__(self):
        return f"Integer({self.data},{self.endianess},{self.size},{self.signed})"

class S(Integer):
    def __init__(self, data, e='be', size=1):
        endian = "big" if e == 'be' else 'little'
        Integer.__init__(self, data, endian, size, True)


class U(Integer):
    def __init__(self, data, e='be', size=1):
        endian = "big" if e == 'be' else 'little'
        Integer.__init__(self, data, endian, size, False)



class Link:
    def __init__(self, hex_data, bytes_data):
        self.hex_data = hex_data
        self.bytes_data = bytes_data

    def get_bytes(self):
        return self.bytes_data

    def get_hex_dump(self):
        return self.hex_data

    def get_writable(self, idx=0):
        if idx == 0:
            return self.get_bytes()
        if idx == 1:
            return self.get_hex_dump()

    def __add__(self, other):
        return Link(self.hex_data + " " + other.get_hex_dump(), self.bytes_data + other.get_bytes())

    def __str__(self):
        return self.hex_data

    def __repr__(self):
        return self.hex_data, str(self.bytes_data)


class HexLink:
    def __init__(self, hex_data):
        self.hex_data = hex_data

    def get_hex_dump(self):
        return self.hex_data

    def get_writable(self):
        return self.hex_data

    def __and__(self, other):
        return HexLink(self.hex_data + " " + other.get_hex_dump())

    def __str__(self):
        return self.hex_data

    def __repr__(self):
        return self.hex_data


class ByteLink:
    def __init__(self, bytes_data):
        self.bytes_data = bytes_data

    def get_bytes(self):
        return self.bytes_data

    def get_writable(self):
        return self.get_bytes()

    def __or__(self, other):
        return HexLink(self.bytes_data + other.get_bytes())

    def __str__(self):
        return str(self.bytes_data)

    def __repr__(self):
        return str(self.bytes_data)



from datatype_writer.format import Format


class ByteArray(Format):
    def __init__(self, data, endianess='be', size=1):
        Format.__init__(self, data, endianess)
        self.size = size

    def get_bytes(self):
        if self.endianess == "le":
            return self.data[::-1]
        else:
            return self.data

    def get_hex_dump(self):
        byte_array = self.get_bytes()
        hex_dump = [hex(b)[2:].zfill(2) for b in byte_array]
        return " ".join(hex_dump).upper()

    def __repr__(self):
        return f"ByteArray({self.data},{self.endianess},{self.size})"


from setuptools import setup

VERSION = '1.0.0'
DESCRIPTION = 'Framework for writing binary data in python'
LONG_DESCRIPTION = 'Framework that allows the use of operatators, such as & or |, for chaining different datatypes together and allowing for easy writing.'

setup(
    name="datatype_writer",
    version=VERSION,
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    author="Ada L",
    author_email="the.nostra.tymus@gmail.com",
    license='MIT',
    packages=['datatype_writer'],
    install_requires=[
        'pyyaml',
    ],
    keywords='machine learning',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3",
    ],
    include_package_data=True,
)



from datatype_writer.integer import Integer
from datatype_writer.string import String
from datatype_writer.bit import Bit
from datatype_writer.byte_array import ByteArray

import yaml
import os
import random
import re


def get_max(size, signed):
    if signed:
        return int((2 ** (8 * size)) / 2 - 1)
    else:
        return int((2 ** (8 * size)) - 1)


def get_min(size, signed):
    if signed:
        return -1 * int((2 ** (8 * size)) / 2)
    else:
        return 0


def get_random_byte_array(size):
    ba = b''
    for _ in range(size):
        ba += int(random.randint(0, 255)).to_bytes(length=1, byteorder='big')
    return ba


def get_random_bits(size):
    bits = ""
    for _ in range(size):
        bits += '10'[random.randint(0, 1)]
    return bits


def get_random_string(count):
    letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890'
    rtn_str = ""
    for _ in range(count):
        rtn_str += letters[random.randint(0, len(letters) - 1)]


def get_random_int(contraint, size, signed):
    min = None
    max = None
    must_be = None
    any_of = None
    none_of = None
    if contraint is not None:
        if type(contraint) == int:
            must_be = contraint
        elif 'eq' in contraint.keys():
            must_be = contraint['eq']
        if 'max' in contraint.keys():
            max = contraint['max']
        if 'min' in contraint.keys():
            min = contraint['min']
        if 'any-of' in contraint.keys():
            any_of = contraint['any-of']
        if 'none-of' in contraint.keys():
            none_of = contraint['none_of']
    if min is not None and max is not None:
        return random.randint(min, max)
    if min is not None:
        return random.randint(min, get_max(size, signed))
    if max is not None:
        return random.randint(get_min(size, signed), max)
    if must_be is not None:
        return must_be
    if any_of is not None:
        return any_of[random.randint(0, len(any_of) - 1)]
    if none_of is not None:
        while True:
            candidate = random.randint(get_min(size, signed), get_max(size, signed))
            if candidate not in none_of:
                return candidate

    return random.randint(get_min(size, signed), get_max(size, signed))


def read_spec(yaml_path):
    if not os.path.exists(yaml_path):
        raise Exception(f"{yaml_path} not found.")

    with open(yaml_path) as file_in:
        return yaml.safe_load(file_in)


def get_size(t):
    i = re.search(r"\d+", t)
    if i is not None:
        return i.group(0)
    else:
        # TODO: find the default
        return 1


def get_endian(t):
    i = re.search(r"le", t)
    if i is not None:
        return "le"
    i = re.search(r"be", t)
    if i is not None:
        return "be"
    return None


class Specification:
    def __init__(self, yaml_path):
        self.spec = read_spec(yaml_path)
        self.enums = None
        self.types = None
        self.default_endian = None
        self.fields = None
        self.process_spec()

    def get_spec(self):
        return self.spec

    def get_repr(self):
        # return the repr of the individual components
        pass

    def get_spec_fields(self):
        return self.fields

    def add_field(self, field):
        self.fields.append(field)

    def field_count(self):
        return len(self.fields)

    def process_spec(self):
        self.default_endian = self.spec['meta']['endian']
        self.enums = self.spec['enums']
        self.types = self.spec['types']
        self.process_seq(self.spec['seq'])

    def get_field_value(self, key):
        for f in self.fields:
            if f.id == key:
                if f.content.data is None:
                    # call self generate data
                    return f.content.data
                else:
                    return f.content.data

    def get_field(self, key):
        for f in self.fields:
            if f.id == key:
                return f

    def process_seq(self, seq):
        for field in seq:
            if 'contains' in field.keys():
                # TODO Add processing for contains constant types
                pass

            elif 'type' in field.keys():
                if type(field['type']) == dict:
                    # If the type is a switch case, choose a value for the reference field then
                    # process the appropriate sequence.
                    field_id = field['type']['switch-on']
                    values = field['type']['cases'].keys()
                    choice = random.randint(0, len(values) - 1)
                    case_value = values[choice]
                    if choice == (len(values) - 1):
                        ref_field = self.get_field(field_id)
                        ref_field.set_constraint({'none-of': [v for v in values[:-1]]})
                        ref_field.assign_value()
                    else:
                        ref_field = self.get_field(field_id)
                        ref_field.content.set_data(case_value)

                    seq_type = field['type']['cases'][case_value]
                    self.process_seq(self.types[seq_type])

                elif field['type'] in self.types.keys(): # The type is a named sequence
                    self.process_seq(self.types[seq['type']])
                else: # The type is an atomic type of int, bit, or str
                    self.process_atomic_field(field)
            else:
                self.process_byte_array(field)

    def process_atomic_field(self, field):
        temp_field = Field()
        temp_field.set_id(field['id'] if 'id' in field.keys() else "")
        t = field['type']
        temp_field.set_type(t)
        if t.find('str') != 0:
            temp_field.set_content(self.process_atomic_string(field))
        elif t[0] == 'u':
            temp_field.set_content(self.process_atomic_signed_int(field, False))
        elif t[0] == 's':
            temp_field.set_content(self.process_atomic_signed_int(field, True))
        elif t[0] == 'b':
            temp_field.set_content(self.process_atomic_bit(field))

        if 'valid' in field.keys():
            temp_field.set_constraint(field['valid'])
        temp_field.set_position(self.field_count())
        temp_field.assign_value()
        self.add_field(temp_field)

    def process_atomic_string(self, field):
        if type(field['size']) == int:
            size = field['size']
        else:
            size = self.get_field_value(key=field['size'])
        if 'encoding' in field.keys():
            encoding = field['encoding'].lower()
        else:
            encoding = 'utf-8'
        return String(None, self.default_endian, size, encoding)

    def process_atomic_signed_int(self, field, signed):
        size = get_size(field['type'])
        endian = get_endian(field['type']) if get_endian(field['type']) is not None else self.default_endian
        return Integer(None, size, endian, signed=signed)

    def process_atomic_bit(self, field):
        size = get_size(field['type'])
        return Bit(None, size)

    def process_byte_array(self, field):
        temp_field = Field()
        temp_field.set_id(field['id'] if 'id' in field.keys() else "")
        temp_field.set_type("BYTES")

        if type(field['size']) == int:
            size = field['size']
        else:
            size = self.get_field_value(key=field['size'])

        content = ByteArray(None, endianess=self.default_endian, size=size)
        temp_field.set_content(content)
        temp_field.set_position(self.field_count())
        self.add_field(temp_field)

    def write_spec_to_file(self, file):
        receipt = []
        bit_buffer = None
        with open(file, "wb") as file_out:
            for field in self.fields:
                if type(field) == Bit:
                    if bit_buffer is None:
                        bit_buffer = field
                    else:
                        bit_buffer = bit_buffer + field
                else:
                    if bit_buffer is not None:
                        file_out.write(bit_buffer.get_bytes())
                        receipt.append(bit_buffer.__repr__())
                        bit_buffer = None
                    file_out.write(field.get_bytes())
                    receipt.append(field.__repr__())
        return receipt

    def get_hex_dump(self):
        dumps = []
        bit_buffer = None
        for field in self.fields:
            if type(field) == Bit:
                if bit_buffer is None:
                    bit_buffer = field
                else:
                    bit_buffer = bit_buffer + field
            else:
                if bit_buffer is not None:
                    dumps.append(bit_buffer.get_hex_dump())
                    bit_buffer = None
                dumps.append(field.get_hex_dump())
        return " ".join(dumps)

class Field:
    def __init__(self):
        self.content = None
        self.constraint = None
        self.position = None
        self.id = None
        self.type_literal = None
        self.enum = None
        self.exclusions = None

    def set_id(self, value):
        self.id = value

    def set_type(self, value):
        self.type_literal = value

    def set_content(self, content):
        self.content = content

    def set_constraint(self, constraint):
        self.constraint = constraint

    def set_position(self, pos):
        self.position = pos

    def update_content_value(self, value):
        self.content.data = value

    def assign_value(self):
        size = self.content.size
        if type(self.content) == String:
            self.content.set_data(get_random_string(size))
        elif type(self.content) == Integer:
            signed = self.content.signed
            self.content.set_data(get_random_int(self.contraint, size, signed))
        elif type(self.content) == ByteArray:
            self.content.set_data(get_random_byte_array(size))
        elif type(self.content) == Bit:
            self.content.set_data(get_random_bits(size))
